const AWSLambdaCaller = require('./services/AWSLambdaCaller');
const Util = require('./util');
const MediaResolver = require('./util/MediaResolver');
const DialogFlowAccessor = require('./services/DialogFlowAccessor');
const KeyMatcherAccessor = require('./services/KeyMatcherAccessor');
const RasaAccessor = require('./services/RasaAccessor');
const defaultFallbackIntent = require('./util/defaultFallbackIntent');
const RequestPreprocessor = require('./util/RequestPreprocessor');
const _ = require('lodash');
const md5 = require('md5');
const IntentResolver = require('./util/IntentResolver');
const AsksuiteTranslator = require('./util/Translator');

const DEFAULT_LANGUAGE = 'pt-br';

class AsksuiteProcessingNaturalUtils {

    constructor(config) {
        this.config = config;

        if (!_.isEmpty(_.get(this.config, 'redis'))) {
            this.redis = this.config.redis;
        }
    }

    withRedis() {
        return !_.isEmpty(_.get(this.config, 'redis'));
    }

    resolveLanguage(paramLanguage) {
        const language = paramLanguage.toLowerCase();

        if (language.startsWith('pt')) {
            return DEFAULT_LANGUAGE;
        } else if (language.startsWith('en')) {
            return 'en';
        } else if (language.startsWith('es')) {
            return 'es';
        }

        throw new Error('language not defined');
    }

    async getInCacheDialogflow(request) {
        if (!this.withRedis()) {
            return Promise.resolve(null);
        }

        const hash = this.generateHash(request);

        const language = this.resolveLanguage(request.language);

        const data = await this.getInCache(`dialogflow:${language}:${hash}`);

        if (!data) {
            return Promise.resolve(null);
        }

        const {intent, extractedData} = data;
        console.log(`OBTAINED BY CACHE ${intent}`);

        return {intent, extractedData};
    }

    findDialogByIntent(request, intent) {
        const {intents, companyId, text} = request;

        const intentResolver = new IntentResolver(intents, companyId, text);
        return intentResolver.findDialogByIntent(intent);
    }

    generateHash(request) {
        const {text} = request;

        const plainText = _.chain(text)
            .deburr() // special characters
            .toLower() // lowercase
            .value() // get value from chain
            .replace(/\s{2,}/g, ' ') // remove duplicate spaces
            .replace(/[^\w\/@ ]/g, ''); // remove all characters dont in (word, /, @, space)

        return md5(plainText);
    }

    async saveInCacheDialogflow(intent, request) {
        if (!this.withRedis()) {
            return Promise.resolve(null);
        }

        if (Util.isNaoentendi(intent.intent)) {
            return Promise.resolve(null);
        }

        const {text} = request;
        const hash = this.generateHash(request);

        const language = this.resolveLanguage(request.language);

        return this.setInCache(`dialogflow:${language}:${hash}`, {
            intent: intent.intent,
            extractedData: intent.extractedData,
            text,
            hash,
        });
    }

    async getInCache(key) {
        return this.redis.getValue(`${key}`);
    }

    async setInCache(key, intent) {
        return this.redis.setValueWithExpirationTime(`${key}`, intent, true, AsksuiteProcessingNatural.EXPIRATION_TIME);
    }

    async translateRequest(request) {
        if (this.resolveLanguage(request.language) !== DEFAULT_LANGUAGE && !request.translation) {
            request.translated = true;

            const asksuiteTranslator = new AsksuiteTranslator(request.config.INFOCHAT_DATA_SERVER, request.config.GOOGLE_TRANSLATE_KEY);
            const {translation} = await asksuiteTranslator.googleTranslateIfNeeded(request.text, request.language);

            request.translation = translation;
        }
    }

    async copyRequestToDefaultLanguage(request) {

        await this.translateRequest(request);

        const toSend = _.clone(request);
        toSend.text = toSend.translation;
        toSend.language = DEFAULT_LANGUAGE;

        return RequestPreprocessor.process(toSend);
    }

    async findIntentDialogFlow(dialogFlowAccessor, request, toDefaultLanguage) {
        if (toDefaultLanguage) {
            await this.translateRequest(request);
        }

        let fromCache = false;

        let toSend = request;
        if (toDefaultLanguage) {
            toSend = await this.copyRequestToDefaultLanguage(request);
        }

        let intent = await this.getInCacheDialogflow(toSend);
        if (!intent) {
            intent = await dialogFlowAccessor.resolveText(toSend);
            await this.saveInCacheDialogflow(intent, toSend);
        } else {
            fromCache = true;
        }
        const extractedData = intent.extractedData;
        intent = this.findDialogByIntent(toSend, intent.intent);

        intent.fromCache = fromCache;
        intent.extractedData = extractedData;

        return intent;
    }
}

class AsksuiteProcessingNatural {
    static get ORDERED_ACCESSORS() {
        return ['DIALOG_FLOW', 'RASA', 'KEY_MATCHER'];
    } // Robot will search for intents in this order default
    static get RASA_FIRST_ORDERED_ACCESSORS() {
        return ['RASA', 'DIALOG_FLOW', 'KEY_MATCHER'];
    } // Robot will search for intents in this order if tryRasaFirst option is set to true
    static get EXPIRATION_TIME() {
        return 60 * 60 * 24;
    } // 24 hrs

    constructor(config) {
        this.config = config;
        this.utils = new AsksuiteProcessingNaturalUtils(this.config);
    }

    findIntent(request) {

        return new Promise((resolve, reject) => {

            const executor = async (request) => {
                let intent = {
                    fromCache: false,
                };

                request = await RequestPreprocessor.process(request);

                const awsLambdaCaller = new AWSLambdaCaller(request.config.CORE_LAMBDA_AWS);
                const dialogFlowAccessor = new DialogFlowAccessor(awsLambdaCaller);
                const keyMatcherAccessor = new KeyMatcherAccessor(awsLambdaCaller);
                const rasaAccessor = new RasaAccessor(request.config.RASA_URL, request.config.RASA_CONFIDENCE_THRESHOLD);

                let orderOption = AsksuiteProcessingNatural.ORDERED_ACCESSORS;
                if (request.tryRasaFirst) {
                    orderOption = AsksuiteProcessingNatural.RASA_FIRST_ORDERED_ACCESSORS;
                }

                request.findIntent = defaultFallbackIntent(request.companyId);
                request.translated = false;
                request.translation = null;

                for (const accessorName of orderOption) {
                    switch (accessorName) {
                    case 'DIALOG_FLOW':
                        intent = await this.utils.findIntentDialogFlow(dialogFlowAccessor, request, false);

                        if (Util.isNaoentendi(intent.intent) && this.utils.resolveLanguage(request.language) !== DEFAULT_LANGUAGE) {
                            intent = await this.utils.findIntentDialogFlow(dialogFlowAccessor, request, true);
                        }

                        break;
                    case 'KEY_MATCHER':
                        intent = await keyMatcherAccessor.resolveText(request);

                        if (Util.isNaoentendi(intent.intent) && this.utils.resolveLanguage(request.language) !== DEFAULT_LANGUAGE) {
                            const toSend = await this.utils.copyRequestToDefaultLanguage(request);
                            intent = await keyMatcherAccessor.resolveText(toSend);
                        }

                        intent.keywordTry = {try: true};
                        break;
                    case 'RASA':
                        intent = await rasaAccessor.resolveText(request);
                        break;
                    }

                    request.findIntent = intent;

                    if (intent.intent && !Util.isNaoentendi(intent.intent)) {
                        intent.resolver = accessorName;

                        if (intent.resolver === 'KEY_MATCHER') {
                            intent.keywordTry.match = intent.intent;
                            intent.keywordTry.intent = intent.intent;
                        }

                        //FOR exit condition
                        break;
                    }

                }

                intent.translated = request.translated;
                intent.translation = request.translation;

                return intent;

            };

            executor(request).then(resolve).catch(reject);

        });

    }

    resolveMedia(request) {
        return new Promise((resolve, reject) => {
            try {
                const mediaResolver = new MediaResolver(request.companyId);
                mediaResolver.resolve(request.media, request.intents).then(resolve).catch(reject);
            } catch (e) {
                reject(e);
            }
        });
    }
}

module.exports = AsksuiteProcessingNatural;
